/*
    This file is part of Corrade.

    Copyright © 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
                2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
              Vladimír Vondruš <mosra@centrum.cz>

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included
    in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

#include <vector>

#include "Corrade/Containers/ArrayViewStl.h"
#include "Corrade/Containers/StringIterable.h"
#include "Corrade/Containers/StringStlView.h"
#include "Corrade/TestSuite/Tester.h"
#include "Corrade/TestSuite/Compare/Container.h"
#include "Corrade/Utility/DebugStlStringView.h"

namespace Corrade { namespace Containers { namespace Test { namespace {

struct StringStlViewTest: TestSuite::Tester {
    explicit StringStlViewTest();

    void convertToStlStringView();
    void convertToStlStringViewEmpty();
    void convertFromStlStringView();
    void convertFromStlStringViewEmpty();

    void convertViewToStlStringView();
    void convertViewToStlStringViewEmpty();
    void convertMutableViewToStlStringView();
    void convertMutableViewToStlStringViewEmpty();
    void convertViewFromStlStringView();
    void convertViewFromStlStringViewEmpty();

    void convertMutableViewFromStlStringView();

    void iterable();
    void iterableVector();
};

StringStlViewTest::StringStlViewTest() {
    addTests({&StringStlViewTest::convertToStlStringView,
              &StringStlViewTest::convertToStlStringViewEmpty,
              &StringStlViewTest::convertFromStlStringView,
              &StringStlViewTest::convertFromStlStringViewEmpty,

              &StringStlViewTest::convertViewToStlStringView,
              &StringStlViewTest::convertViewToStlStringViewEmpty,
              &StringStlViewTest::convertMutableViewToStlStringView,
              &StringStlViewTest::convertMutableViewToStlStringViewEmpty,
              &StringStlViewTest::convertViewFromStlStringView,
              &StringStlViewTest::convertViewFromStlStringViewEmpty,

              &StringStlViewTest::convertMutableViewFromStlStringView,

              &StringStlViewTest::iterable,
              &StringStlViewTest::iterableVector});
}

using namespace Literals;
using namespace std::string_view_literals;

void StringStlViewTest::convertToStlStringView() {
    String a = "hello\0!!!"_s;
    std::string_view b = a;
    CORRADE_COMPARE(b, "hello\0!!!"sv);
    CORRADE_COMPARE(b.data(), static_cast<const void*>(a.data()));
}

void StringStlViewTest::convertToStlStringViewEmpty() {
    String a;
    std::string_view b = a;
    CORRADE_COMPARE(b, std::string_view{});
    CORRADE_COMPARE(b.data(), static_cast<const void*>(a.data()));
}

void StringStlViewTest::convertFromStlStringView() {
    const std::string_view a = "hello\0!!!"sv;
    String b = a;
    CORRADE_COMPARE(b, "hello\0!!!"_s);
}

void StringStlViewTest::convertFromStlStringViewEmpty() {
    const std::string_view a;
    String b = a;
    CORRADE_COMPARE(b, ""_s);
}

void StringStlViewTest::convertViewToStlStringView() {
    StringView a = "hello\0!!!"_s;
    std::string_view b = a;
    CORRADE_COMPARE(b, "hello\0!!!"sv);
    CORRADE_COMPARE(b.data(), static_cast<const void*>(a.data()));
}

void StringStlViewTest::convertViewToStlStringViewEmpty() {
    StringView a;
    std::string_view b = a;
    CORRADE_COMPARE(b, std::string_view{});
    CORRADE_COMPARE(b.data(), static_cast<const void*>(nullptr));
}

void StringStlViewTest::convertMutableViewToStlStringView() {
    char data[] = "hello\0!!!";
    MutableStringView a{data, 9};
    std::string_view b = a;
    CORRADE_COMPARE(b, "hello\0!!!"sv);
    CORRADE_COMPARE(b.data(), static_cast<const void*>(data));
}

void StringStlViewTest::convertMutableViewToStlStringViewEmpty() {
    MutableStringView a;
    std::string_view b = a;
    CORRADE_COMPARE(b, std::string_view{});
    CORRADE_COMPARE(b.data(), static_cast<const void*>(nullptr));
}

void StringStlViewTest::convertViewFromStlStringView() {
    const std::string_view a = "hello\0!!!"sv;
    StringView b = a;
    CORRADE_COMPARE(b, "hello\0!!!"_s);
    CORRADE_COMPARE(b.data(), static_cast<const void*>(a.data()));
}

void StringStlViewTest::convertViewFromStlStringViewEmpty() {
    const std::string_view a = "hello\0!!!"sv;
    StringView b = a;
    CORRADE_COMPARE(b, "hello\0!!!"_s);
    CORRADE_COMPARE(b.data(), static_cast<const void*>(a.data()));
}

void StringStlViewTest::convertMutableViewFromStlStringView() {
    /* A string_view should never be convertible to a mutable view. Not using
       is_convertible to catch also accidental explicit conversions. */
    CORRADE_VERIFY(std::is_constructible<StringView, std::string_view>::value);
    CORRADE_VERIFY(!std::is_constructible<MutableStringView, std::string_view>::value);
}

void StringStlViewTest::iterable() {
    std::string_view data[]{"hello", "world", "!"};
    ArrayView<std::string_view> a = data;

    StringIterable ai = a;
    CORRADE_COMPARE_AS(ai, (StringIterable{
        "hello", "world", "!"
    }), TestSuite::Compare::Container);
}

void StringStlViewTest::iterableVector() {
    /* Slightly better but still awful! */
    std::vector<std::string_view> a{"hello", "world", "!"};

    StringIterable ai = a;
    CORRADE_COMPARE_AS(ai, (StringIterable{
        "hello", "world", "!"
    }), TestSuite::Compare::Container);
}

}}}}

CORRADE_TEST_MAIN(Corrade::Containers::Test::StringStlViewTest)
